#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "SDL.h"
#include "SDL_mmap.h"
#include "charactron.h"
#include "bsicon.h"
#include "font_5x9.h"

#define REALLOCN(PTR,TYPE,N) ((TYPE *)realloc(PTR,sizeof(TYPE)*(N)))

int render_page(charactron *c, const SDL_mspan *m, 
                mempos_t *start, mempos_t *end);
mempos_t *parse_mmap_text(charactron *c, const SDL_mspan *m);

static const char *filename=NULL;

void setcaptionf(const char *cmdstr, ...)
{
  char buf[1024];
  va_list ap;

  va_start(ap,cmdstr);
    vsprintf(buf,cmdstr,ap);
  va_end(ap);

  SDL_WM_SetCaption(buf,"Reader");
}

void browse_pages(charactron *c, const mempos_t *pages, const SDL_mspan *m)
{
  SDL_Event event;
  int page=1,running=1,redraw=1;

  while(running)
  {
    if(redraw)
    {
      mempos_t p=pages[page];
      render_page(c,m,&p,NULL);
      setcaptionf("%s [%d/%ld]",filename,page,pages[0]);
      SDL_Flip(SDL_GetVideoSurface());
      redraw=0;
    }

    SDL_WaitEvent(&event);
    switch(event.type)
    {
    Uint8 mod;

    case SDL_KEYDOWN:
      mod=SDL_GetModState();
      redraw=1;

      switch(event.key.keysym.sym)
      {
      case SDLK_RIGHT:
      case SDLK_PAGEDOWN:
      case SDLK_SPACE:
        if(mod & KMOD_CTRL)
          page+=10;
        else
          page++;

        if(page > pages[0]) page=pages[0];
        break;

      case SDLK_LEFT:
      case SDLK_PAGEUP:
      case SDLK_BACKSPACE:
        if(mod & KMOD_CTRL)
          page-=10;
        else
          page--;

        if(page < 1) page=1;
        break;

      case SDLK_HOME:
        page=1;
        break;

      case SDLK_END:
        page=pages[0];
        break;

      case SDLK_ESCAPE:
        running=0;
        break;

      default:
        break;
      }
      break;
    case SDL_QUIT:
      running=0;
      break;

    default:
      break;
    }
  }
}

#ifdef _WIN32_WCE
#define DEFAULT_W	240
#define DEFAULT_H	(320-(8*7))
#else
#define DEFAULT_W	(5*80)
#define DEFAULT_H	(9*25)
#endif

static const SDL_Color black={0,0,0,0};
static const SDL_Color white={255,255,255,0};

int scan_color(const char *cstr, SDL_Color *out)
{
  if(sscanf(cstr,"%02hhx%02hhx%02hhx\n",
    &(out->r),
    &(out->g),
    &(out->b))<3)
    return(-1);

  out->unused=0;
  return(0);
}

int main(int argc, char *argv[])
{
  int n;
  charactron *c=NULL;

  if(SDL_Init(SDL_INIT_VIDEO)<0)
  {
    fprintf(stderr,"Couldn't init SDL: %s\n",SDL_GetError());
    return(1);
  }

  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

  BS_SetIcon();

  if(SDL_SetVideoMode(DEFAULT_W,DEFAULT_H,16,SDL_ANYFORMAT)==NULL)
  {
    fprintf(stderr,"Couldn't set video mode: %s\n",SDL_GetError());
    goto MAIN_FAIL;
  }

  c=charactron_load(get_font_5x9(),NULL,1);
  if(c==NULL)
  {
    fprintf(stderr,"Couldn't open charactron with default font\n");
    goto MAIN_FAIL;
  }

  for(n=1; n<argc; n++)
  {
    mempos_t *pages=NULL;
    mmap_t len;
    SDL_mspan *mem=NULL;

    struct SDL_mmap *mm;

    if((argc-1) > n)
    {
      SDL_Color cl;
      if(strcmp(argv[n],"-fg")==0)
      {
        n++;
        if(scan_color(argv[n],&cl)<0)
          fprintf(stderr,"Invalid color \"%s\"\n",argv[n]);

        charactron_set_color(c,CHARACTRON_FG,cl);
        continue;
      }
      else if(strcmp(argv[n],"-bg")==0)
      {
        n++;
        if(scan_color(argv[n],&cl)<0)
          fprintf(stderr,"Invalid color \"%s\"\n",argv[n]);

        charactron_set_color(c,CHARACTRON_BG,cl);
        continue;
      }
    }

    mm=SDL_mmap_open(argv[n], SDL_MMAP_RD, SDL_MADVISE_SEQUENTIAL);
    if(mm==NULL)
    {
      fprintf(stderr,"Couldn't mmap %s (%p): %s\n",argv[n],mm,SDL_GetError());
      continue;
    }

    if(SDL_mmap_len(mm,&len)<0)
    {
      fprintf(stderr,"Couldn't read length of file\n");
      goto MAIN_NEXT;
    }

    if(SDL_mmap_maxchunk() < len)
    {
      fprintf(stderr,"Memory model is too small to map this file");
      goto MAIN_NEXT;
    }

    mem=SDL_mmap(mm,(mempos_t)len,SDL_MMAP_RD,0);
    if(mem==NULL)
    {
      fprintf(stderr,"Couldn't map into memory\n");
      goto MAIN_NEXT;
    }

    pages=parse_mmap_text(c,mem);
    if(pages == NULL)
    {
      fprintf(stderr,"Couldn't pageify text!\n");
      goto MAIN_NEXT;
    }

    fprintf(stderr,"Book has %ld pages\n",pages[0]);

	{
      const char *next;
      filename=argv[n];
	  next=strpbrk(filename,"/\\");
      while(next != NULL)
	  {
        filename=next+1;
        next=strpbrk(filename,"/\\");
	  }
	}

    browse_pages(c,pages,mem);

MAIN_NEXT:
    if(pages != NULL)	free(pages);
    if(mem != NULL)	SDL_munmap(mem);
    if(mm != NULL)	SDL_mmap_close(mm);
  }

  charactron_free(c);
  SDL_Quit();
  return(0);

MAIN_FAIL:
  if(c != NULL) charactron_free(c);
  SDL_Quit();
  return(1);
}

int render_page(charactron *c, const SDL_mspan *m, 
                mempos_t *start, mempos_t *end)
{
  mempos_t pos,max,wordstart=0;
  enum
  {
    STATE_START,
    STATE_WORD,
    STATE_SPACE
  } state=STATE_START;

  charactron_formfeed(c);

  if(end == NULL) max=m->len;
  else            max=(*end);

  if(start == NULL) pos=0;
  else              pos=(*start);

  while(pos < max)
  {
    switch(state)
    {
    case STATE_START:
      if(isspace(m->mem[pos]))
        break;

      wordstart=pos;
      state=STATE_WORD;

      if(start != NULL)
        (*start)=pos;

    case STATE_WORD:
      if(!isspace(m->mem[pos]))
        break;

      if((pos-wordstart) > c->maxx)
      {
        while(wordstart < pos)
        {
          int len=(pos-wordstart);

          if(c->x > 0)
            if(charactron_linefeed(c)<0)
            {
              pos=wordstart;
              goto RENDER_PAGE_END;
            }

          if(len > CHARACTRON_W_FREE(c))
            len=CHARACTRON_W_FREE(c);

          charactron_draw_data(c,m->mem+wordstart,len);
          wordstart+=len;
        }
      }
      else while(charactron_draw_data(c,m->mem+wordstart,(pos-wordstart))<0)
      {
        if(charactron_linefeed(c)<0)
          goto RENDER_PAGE_END;
      }

      wordstart=pos;
      state=STATE_SPACE;

    case STATE_SPACE:
      switch(m->mem[pos])
      {
      case ' ':
      case '\t':
        if(CHARACTRON_W_FREE(c) > 0)
          c->x++;
        else if(charactron_linefeed(c)<0)
        {
          wordstart=pos;
          goto RENDER_PAGE_END;
        }
        break;

      case '\r':
        c->x=0;
        break;

      case '\n':
        if(charactron_linefeed(c)<0)
        {
          wordstart=pos;
          goto RENDER_PAGE_END;
        }
        break;

      case '\f':
        wordstart=pos;
        goto RENDER_PAGE_END;
        break;

      default:
        wordstart=pos;
        pos--;
        state=STATE_WORD;
        break;
      }
      break;
    }

    pos++;
  }

  switch(state)
  {
  case STATE_WORD:
      if((pos-wordstart) > c->maxx)
      {
        while(wordstart < pos)
        {
          int len=(pos-wordstart);

          if(c->x > 0)
            if(charactron_linefeed(c)<0)
            {
              pos=wordstart;
              goto RENDER_PAGE_END;
            }

          if(len > CHARACTRON_W_FREE(c))
            len=CHARACTRON_W_FREE(c);

          charactron_draw_data(c,m->mem+wordstart,len);
          wordstart+=len;
        }
      }
      else while(charactron_draw_data(c,m->mem+wordstart,(pos-wordstart))<0)
      {
        if(charactron_linefeed(c)<0)
          goto RENDER_PAGE_END;
      }

      wordstart=pos;

    break;
  case STATE_START:
  case STATE_SPACE:
    wordstart=pos;
    break;
  default:
    break;
  }

RENDER_PAGE_END:
  if(end != NULL) (*end)=wordstart;

//  if((end != NULL) && (start != NULL))
//    fprintf(stderr,"render_page:  %ld - %ld\n",*start,*end);
  return(0);
}

mempos_t *parse_mmap_text(charactron *c, const SDL_mspan *m)
{
  Uint32 oldflag=c->flag;
  mempos_t *pages=NULL;
  int numpages=0;
  mempos_t pos=(mempos_t)0;

  CHARACTRON_DISABLE(c);

  while(pos < m->len)
  {
    mempos_t npos=pos,nend=m->len;

    render_page(c,m,&npos,&nend);

    if(pos > nend)// || (nend == 0L))
      break;
    pos=nend;

    pages=REALLOCN(pages,mempos_t,((++numpages) + 1));
    pages[numpages]=npos;
  }

  pages[0]=numpages;
  c->flag=oldflag;
  return(pages);
}
